import Solid, { ignore, Core, Utils, root } from 'solid-js';

function map(fn) {
  return function(obsv) {
    obsv = Solid(obsv);
    return Solid(() => {
      var v;
      if ((v = obsv.value) === void 0) {
        return;
      }
      return ignore(() => {
        return fn(v);
      });
    }, {
      notifyAlways: true
    });
  };
}

function mapS(mapFn) {
  return function(obsv) {
    var disposables, length, list, mapped, signals, trackIndex;
    mapped = [];
    list = [];
    disposables = [];
    length = 0;
    if ((trackIndex = mapFn.length > 1)) {
      signals = [];
    }
    Core.context.disposables.push(function() {
      var d, k, len;
      for (k = 0, len = disposables.length; k < len; k++) {
        d = disposables[k];
        d();
      }
      return disposables = null;
    });
    return map((newList) => {
      var d, end, i, indexedItems, item, itemIndex, j, k, l, len, len1, len2, m, newEnd, newLength, newListUnwrapped, newMapped, start, tempDisposables, tempSignals;
      // non-arrays
      newListUnwrapped = Utils.unwrap(newList, true);
      if (!Array.isArray(newListUnwrapped)) {
        if ((newListUnwrapped == null) || newListUnwrapped === false) {
          mapped = [];
          list = [];
          for (k = 0, len = disposables.length; k < len; k++) {
            d = disposables[k];
            d();
          }
          disposables = [];
          if (trackIndex) {
            signals = [];
          }
          return null;
        }
        if (list[0] === newListUnwrapped) {
          return mapped[0];
        }
        for (l = 0, len1 = disposables.length; l < len1; l++) {
          d = disposables[l];
          d();
        }
        disposables = [];
        list[0] = newListUnwrapped;
        return mapped[0] = root(function(dispose) {
          disposables[0] = dispose;
          return mapFn(newList);
        });
      }
      newLength = newListUnwrapped.length;
      if (newLength === 0) {
        if (length !== 0) {
          list = [];
          mapped = [];
          length = 0;
          for (m = 0, len2 = disposables.length; m < len2; m++) {
            d = disposables[m];
            d();
          }
          disposables = [];
          if (trackIndex) {
            signals = [];
          }
        }
      } else if (length === 0) {
        i = 0;
        while (i < newLength) {
          list[i] = newListUnwrapped[i];
          mapped[i] = root(function(dispose) {
            disposables[i] = dispose;
            if (trackIndex) {
              return mapFn((typeof newList.peek === "function" ? newList.peek(i) : void 0) || newList[i], signals[i] = new Signal(i), newList);
            }
            return mapFn((typeof newList.peek === "function" ? newList.peek(i) : void 0) || newList[i]);
          });
          i++;
        }
        length = newLength;
      } else {
        newMapped = new Array(newLength);
        tempDisposables = new Array(newLength);
        indexedItems = new Map();
        if (trackIndex) {
          tempSignals = new Array(newLength);
        }
        end = Math.min(length, newLength);
        // reduce from both ends
        start = 0;
        while (start < end && newListUnwrapped[start] === list[start]) {
          start++;
        }
        end = length - 1;
        newEnd = newLength - 1;
        while (end >= 0 && newEnd >= 0 && newListUnwrapped[newEnd] === list[end]) {
          newMapped[newEnd] = mapped[end];
          tempDisposables[newEnd] = disposables[newEnd];
          if (trackIndex) {
            tempSignals[newEnd] = signals[newEnd];
          }
          end--;
          newEnd--;
        }
        // create indices
        j = newEnd;
        while (j >= start) {
          item = newListUnwrapped[j];
          itemIndex = indexedItems.get(item);
          if (itemIndex != null) {
            itemIndex.push(j);
          } else {
            indexedItems.set(item, [j]);
          }
          j--;
        }
        // find old items
        i = start;
        while (i <= end) {
          item = list[i];
          itemIndex = indexedItems.get(item);
          if ((itemIndex != null) && itemIndex.length > 0) {
            j = itemIndex.pop();
            newMapped[j] = mapped[i];
            tempDisposables[j] = disposables[i];
            if (trackIndex) {
              tempSignals[j] = signals[i];
            }
          } else {
            disposables[i]();
          }
          i++;
        }
        // set all new values
        j = start;
        while (j < newLength) {
          if (newMapped.hasOwnProperty(j)) {
            mapped[j] = newMapped[j];
            disposables[j] = tempDisposables[j];
            if (trackIndex) {
              signals[j] = tempSignals[j];
              if (signals[j].peek() !== j) {
                signals[j].next(j);
              }
            }
          } else {
            mapped[j] = root(function(dispose) {
              disposables[j] = dispose;
              if (trackIndex) {
                return mapFn((typeof newList.peek === "function" ? newList.peek(j) : void 0) || newList[j], signals[j] = new Signal(j), newList);
              }
              return mapFn((typeof newList.peek === "function" ? newList.peek(j) : void 0) || newList[j]);
            });
          }
          j++;
        }
        // truncate extra length
        length = mapped.length = disposables.length = newLength;
        if (trackIndex) {
          signals.length = newLength;
        }
        // save list for next iteration
        list = newListUnwrapped.slice(0);
      }
      return mapped;
    })(obsv);
  };
}

export { map, mapS };
