import { createRuntime } from 'babel-plugin-jsx-dom-expressions';

var comparer;

comparer = function(v, k, b, is_array, path, r) {
  var new_path, ref, ref1;
  new_path = path.concat([k]);
  if (is_array && !(((v != null ? v.id : void 0) && (v != null ? v.id : void 0) === ((ref = b[k]) != null ? ref.id : void 0)) || ((v != null ? v._id : void 0) && (v != null ? v._id : void 0) === ((ref1 = b[k]) != null ? ref1._id : void 0))) || !((v != null) && ((b != null ? b[k] : void 0) != null) && (v instanceof Object))) {
    return r.push(new_path.concat([v]));
  }
  return r.push.apply(r, diff(v, b[k], new_path));
};

var isObject = function(obj) {
  var ref;
  return obj !== null && ((ref = typeof obj) === 'object' || ref === 'function');
};

var isFunction = function(val) {
  return typeof val === 'function';
};

var diff = function(a, b, path = []) {
  var i, k, l, len, r, v;
  r = [];
  if (!isObject(a) || (b == null)) {
    if (a !== b) {
      r.push(path.concat([a]));
    }
  } else if (Array.isArray(a)) {
    for (k = i = 0, len = a.length; i < len; k = ++i) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v) {
        comparer(v, k, b, true, path, r);
      }
    }
    if ((b != null ? b.length : void 0) > a.length) {
      l = a.length;
      while (l < b.length) {
        r.push(path.concat([l, void 0]));
        l++;
      }
    }
  } else {
    for (k in a) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v) {
        comparer(v, k, b, false, path, r);
      }
    }
    for (k in b) {
      v = b[k];
      if (!(k in a)) {
        r.push(path.concat([k, void 0]));
      }
    }
  }
  return r;
};

var clone = function(v) {
  var k, obj;
  if (!isObject(v)) {
    return v;
  }
  if (Array.isArray(v)) {
    return v.slice(0);
  }
  obj = {};
  for (k in v) {
    obj[k] = v[k];
  }
  return obj;
};

var unwrap = function(item, deep) {
  var k, result, unwrapped, v;
  if (result = item != null ? item._target : void 0) {
    return result;
  }
  if (!(deep && isObject(item) && !isFunction(item) && !(item instanceof Element) && !(item instanceof DocumentFragment))) {
    return item;
  }
  for (k in item) {
    v = item[k];
    if ((unwrapped = unwrap(v, true)) !== v) {
      item[k] = unwrapped;
    }
  }
  return item;
};

var isEqual = function(a, b) {
  return ((a != null ? a._target : void 0) || a) === ((b != null ? b._target : void 0) || b);
};

var Utils = /*#__PURE__*/Object.freeze({
  isObject: isObject,
  isFunction: isFunction,
  diff: diff,
  clone: clone,
  unwrap: unwrap,
  isEqual: isEqual
});

var cancelTask, processUpdates;

cancelTask = function(handle, defer) {
  var index, q;
  q = defer ? Core.queues.deferred : Core.queues.current;
  index = handle - (q.nextHandle - q.tasks.length);
  if (q.nextIndex <= index) {
    return q.tasks[index] = null;
  }
};

processUpdates = function(q, current) {
  var count, err, mark, task;
  count = 0;
  mark = 0;
  while (q.nextIndex < q.tasks.length) {
    if (!(task = q.tasks[q.nextIndex])) {
      q.nextIndex++;
      continue;
    }
    if (current && task.defer) {
      task.handle = void 0;
      queueTask(task, task.value);
      q.nextIndex++;
      continue;
    }
    if (q.nextIndex > mark) {
      if (count++ > 5000) {
        console.error('Exceeded max task recursion');
        q.nextIndex = 0;
        return q.tasks = [];
      }
      mark = q.tasks.length;
    }
    try {
      task(task.value);
      task.handle = task.dhandle = task.value = void 0;
    } catch (error) {
      err = error;
      console.error(err);
    }
    q.nextIndex++;
  }
  q.nextIndex = 0;
  q.tasks = [];
};

var Core = {
  context: null,
  frozen: false,
  queues: {
    deferred: {
      tasks: [],
      nextIndex: 0,
      nextHandle: 1
    }
  },
  clock: 0
};

var queueTask = function(task, value) {
  var q;
  if (Core.frozen) {
    if (task.handle != null) {
      cancelTask(task.handle);
    }
    Core.queues.current.tasks.push(task);
    task.value = value;
    task.handle = Core.queues.current.nextHandle++;
    return;
  }
  if (task.defer) {
    if (task.dhandle != null) {
      cancelTask(task.dhandle, task.defer);
    }
    q = Core.queues.deferred;
    if (!q.tasks.length) {
      Promise.resolve().then(function() {
        processUpdates(q);
        return Core.clock++;
      });
    }
    q.tasks.push(task);
    task.value = value;
    task.dhandle = q.nextHandle++;
    return;
  }
  task(value);
};

var run = function(fn) {
  var execute, prevQueue;
  if (!Core.frozen) {
    execute = true;
    prevQueue = Core.queues.current;
    Core.queues.current = {
      tasks: [],
      nextIndex: 0,
      nextHandle: 1
    };
    Core.frozen = true;
  }
  fn();
  if (execute) {
    Core.frozen = false;
    processUpdates(Core.queues.current, true);
    Core.queues.current = prevQueue;
  }
};

var setContext = function(newContext, fn) {
  var context, ret;
  context = Core.context;
  Core.context = newContext;
  ret = fn();
  Core.context = context;
  return ret;
};

var root = function(fn) {
  var d;
  return setContext({
    disposables: d = []
  }, function() {
    return fn(function() {
      var disposable, i, len;
      for (i = 0, len = d.length; i < len; i++) {
        disposable = d[i];
        disposable();
      }
      d = null;
    });
  });
};

var ignore = function(fn) {
  var disposables;
  ({disposables} = Core.context || {});
  return setContext({disposables}, fn);
};

var onClean = function(fn) {
  if (!Core.context) {
    return;
  }
  return Core.context.disposables.push(fn);
};

var Sync;

var Sync$1 = Sync = class Sync {
  constructor(handler, {defer = true} = {}) {
    var ref;
    this.handler = handler;
    this.exec = this.exec.bind(this);
    this.exec.defer = defer;
    this.disposables = [];
    this.exec();
    if ((ref = Core.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  exec() {
    if (this.__disposed) {
      return;
    }
    this.clean();
    return setContext(this, this.handler);
  }

  clean() {
    var disposable, i, len, ref;
    ref = this.disposables;
    for (i = 0, len = ref.length; i < len; i++) {
      disposable = ref[i];
      disposable();
    }
    return this.disposables = [];
  }

  dispose() {
    if (this.__disposed) {
      return;
    }
    this.clean();
    return this.__disposed = true;
  }

};

function symbolObservablePonyfill(root) {
	var result;
	var Symbol = root.Symbol;

	if (typeof Symbol === 'function') {
		if (Symbol.observable) {
			result = Symbol.observable;
		} else {
			result = Symbol('observable');
			Symbol.observable = result;
		}
	} else {
		result = '@@observable';
	}

	return result;
}

/* global window */

var root$1;

if (typeof self !== 'undefined') {
  root$1 = self;
} else if (typeof window !== 'undefined') {
  root$1 = window;
} else if (typeof global !== 'undefined') {
  root$1 = global;
} else if (typeof module !== 'undefined') {
  root$1 = module;
} else {
  root$1 = Function('return this')();
}

var result = symbolObservablePonyfill(root$1);

var Renderer = createRuntime({
  wrapExpr: function(accessor, fn) {
    return new Sync$1(() => {
      var sub, value;
      value = accessor();
      if (isObject(value)) {
        if ('sid' in value) {
          new Sync$1(() => {
            return fn(value.value);
          });
          return;
        }
        if (result in value) {
          Core.context.disposables.push((sub = value[result]()).subscribe(fn).unsubscribe.bind(sub));
          return;
        }
        if ('then' in value) {
          value.then(fn);
          return;
        }
      }
      return fn(value);
    });
  },
  sanitize: unwrap
});

var S, Selector, Signal, Stream, counter, observableProvider, promiseProvider, resolveAsync, rxCompat, sequenceCounter;

counter = 0;

sequenceCounter = 0;

promiseProvider = function(promise) {
  return function(stream) {
    promise.then(function(value) {
      stream._value = value;
      return stream.notify('next', value);
    }).catch(err)(function() {
      return stream.notify('error', err);
    });
  };
};

observableProvider = function(observable) {
  return function(stream) {
    var disposable;
    disposable = observable.subscribe((value) => {
      stream._value = value;
      return stream.notify('next', value);
    }, (error) => {
      return stream.notify('error', error);
    }, () => {
      return stream.notify('complete');
    });
    return () => {
      return disposable.unsubscribe();
    };
  };
};

resolveAsync = function(value, seq, obsv, fn) {
  if (!isObject(value)) {
    return fn(value);
  }
  if ('subscribe' in value) {
    obsv.disposables.push(value.subscribe(fn));
    return;
  }
  if ('then' in value) {
    value.then(function(value) {
      if (seq < obsv.seq) {
        return;
      }
      return fn(value);
    });
    return;
  }
  return fn(value);
};

rxCompat = function(obsv, op) {
  var newObsv;
  newObsv = {
    subscribe: function(sub) {
      return op.call(sub, obsv);
    },
    lift: function(op) {
      return rxCompat(this, op);
    }
  };
  Object.defineProperty(newObsv, result, {
    value: function() {
      return newObsv;
    }
  });
  return newObsv;
};

Signal = class Signal {
  constructor(value) {
    this.sid = `s_${counter++}`;
    this._subscriptions = new Set();
    this._value = value;
    Object.defineProperty(this, 'value', {
      get: function() {
        var ref;
        if ((ref = Core.context) != null ? ref.exec : void 0) {
          Core.context.disposables.push(this._subscribe(Core.context.exec).unsubscribe);
        }
        return this._value;
      },
      enumerable: true
    });
    Object.defineProperty(this, result, {
      value: function() {
        return this;
      }
    });
  }

  peek() {
    return this._value;
  }

  next(value) {
    this._value = value;
    return this.notify('next', value);
  }

  subscribe(observer) {
    var d;
    d = this._subscribe.apply(this, arguments);
    if (this._value !== void 0) {
      (observer.next || observer)(this._value);
    }
    return d;
  }

  _subscribe(observer) {
    if (typeof observer !== 'object' || observer === null) {
      observer = {
        next: observer,
        error: arguments[1],
        complete: arguments[2]
      };
    }
    this._subscriptions.add(observer);
    return {
      unsubscribe: this.unsubscribe.bind(this, observer)
    };
  }

  unsubscribe(observer) {
    return this._subscriptions.delete(observer);
  }

  notify(type, value) {
    var handler, i, ref, size, sub;
    if (!(size = this._subscriptions.size)) {
      return;
    }
    i = 0;
    ref = this._subscriptions;
    for (sub of ref) {
      if (!(handler = sub[type])) {
        continue;
      }
      queueTask(handler.bind(sub), value);
      if (++i === size) {
        break;
      }
    }
  }

  lift(op) {
    return rxCompat(this, op);
  }

  pipe() {
    return pipe.apply(this, arguments)(this);
  }

};

Stream = class Stream extends Signal {
  constructor(provider) {
    var ref;
    super();
    this.provider = provider;
    if ((ref = Core.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  _subscribe() {
    var disposable, oldSize;
    oldSize = this._subscriptions.size;
    disposable = super._subscribe.apply(this, arguments);
    if (!oldSize && this._subscriptions.size && !this.initialized) {
      this.disposable = this.provider(this);
      this.initialized = true;
    }
    return disposable;
  }

  unsubscribe(fn) {
    var delayed;
    super.unsubscribe(fn);
    delayed = () => {
      if (!this._subscriptions.size) {
        if (typeof this.disposable === "function") {
          this.disposable();
        }
        delete this.disposable;
        return delete this.initialized;
      }
    };
    delayed.defer = true;
    return queueTask(delayed);
  }

  next() {}

  peek() {
    if (!this.initialized) {
      this.disposable = this.provider(this);
      this.initialized = true;
    }
    return super.peek();
  }

  dispose() {
    if (!this.initialized) {
      return;
    }
    if (typeof this.disposable === "function") {
      this.disposable();
    }
    return delete this.initialized;
  }

};

Selector = class Selector extends Signal {
  constructor(handler1, {defer, notifyAlways} = {}) {
    var ref;
    super();
    this.handler = handler1;
    this.notifyAlways = notifyAlways;
    this.exec = this.exec.bind(this);
    this.exec.defer = defer;
    this.disposables = [];
    this.pure = true;
    this.seq = 0;
    if ((ref = Core.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  exec() {
    var context, is_sleeping, seq, value;
    if (this.__disposed) {
      return;
    }
    this.clean();
    is_sleeping = !this._forceAwake && !this._subscriptions.size;
    context = is_sleeping ? null : this;
    value = setContext(context, () => {
      return this.handler(this._value);
    });
    seq = ++sequenceCounter;
    return resolveAsync(value, seq, this, (value) => {
      if (this.__disposed || value === void 0 || (!this.notifyAlways && value === this._value && !isObject(value))) {
        return;
      }
      this.seq = seq;
      return super.next(value);
    });
  }

  next() {}

  peek() {
    if (!this.disposables.length) {
      this.exec();
    }
    return super.peek();
  }

  _subscribe() {
    if (!(this._subscriptions.size || this.disposables.length)) {
      this._forceAwake = true;
      this.exec();
      delete this._forceAwake;
    }
    return super._subscribe.apply(this, arguments);
  }

  unsubscribe(fn) {
    var delayed;
    super.unsubscribe(fn);
    delayed = () => {
      if (!this._subscriptions.size) {
        return this.clean();
      }
    };
    delayed.defer = true;
    return queueTask(delayed);
  }

  clean() {
    var disposable, j, len, ref;
    ref = this.disposables;
    for (j = 0, len = ref.length; j < len; j++) {
      disposable = ref[j];
      disposable();
    }
    return this.disposables = [];
  }

  dispose() {
    if (this.__disposed) {
      return;
    }
    this.notify('complete');
    this.clean();
    return this.__disposed = true;
  }

};

var Signal$1 = S = function(payload, options = {}) {
  if ('sid' in payload) {
    return payload;
  }
  if (isFunction(payload)) {
    return new Selector(payload, options);
  }
  if (result in payload) {
    return new Stream(observableProvider(payload[result]()));
  }
  if ('then' in payload) {
    return new Stream(promiseProvider(payload));
  }
  return new Signal(payload, options);
};

var pipe = function(...fns) {
  if (!fns) {
    return i(() => {
      return i;
    });
  }
  if (fns.length === 1) {
    return fns[0];
  }
  return (obsv) => {
    obsv = S(obsv);
    return fns.reduce(((prev, fn) => {
      return fn(prev);
    }), obsv);
  };
};

export default Signal$1;
export { Utils, Sync$1 as Sync, Renderer as r, pipe, Core, queueTask, run, setContext, root, ignore, onClean };
