'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var Solid = require('solid-js');
var Solid__default = _interopDefault(Solid);

var DEFINED, Handler;

DEFINED = {
  '_target': true,
  'on': true,
  'peek': true
};

var Handler$1 = Handler = class Handler {
  constructor(_target) {
    var i, len, method, ref;
    this._target = _target;
    this._child_subscriptions = {};
    ref = Object.keys(DEFINED).slice(2);
    for (i = 0, len = ref.length; i < len; i++) {
      method = ref[i];
      this[method] = this[method].bind(this);
    }
  }

  get(target, property) {
    var ref, value;
    if (DEFINED[property]) {
      return this[property];
    }
    if (!((ref = Solid.Core.context) != null ? ref.exec : void 0) || property === 'length' || typeof property === 'symbol') {
      return target[property];
    }
    if ((value = target[property]) != null) {
      if (Solid.Utils.isFunction(value)) {
        return value;
      }
      if (Solid.Utils.isObject(value) && Object.isFrozen(value)) {
        value = Solid.Utils.clone(value);
        target[property] = value;
      }
      value = Handler.wrap(value);
    }
    Solid.Core.context.disposables.push(this.on(property, Solid.Core.context.exec).unsubscribe);
    return value;
  }

  set(target, property, value) {
    return true;
  }

  deleteProperty(target, property) {
    return true;
  }

  subscribe(fn) {
    this.__subscriptions || (this.__subscriptions = new Set());
    this.__subscriptions.add(fn);
    return {
      unsubscribe: () => {
        return this.__subscriptions.delete(fn);
      }
    };
  }

  notify(value) {
    var ref, ref1, sub;
    if (!((ref = this.__subscriptions) != null ? ref.size : void 0)) {
      return;
    }
    ref1 = this.__subscriptions;
    for (sub of ref1) {
      Solid.queueTask(sub, value);
    }
  }

};

Handler.wrappers = new WeakMap();

Handler.wrap = function(value) {
  var handler, proxy, wrapper;
  if (!(Solid.Utils.isObject(value) && !(value instanceof Element))) {
    return value;
  }
  if (!(wrapper = Handler.wrappers.get(value))) {
    proxy = new Proxy(value, handler = new Handler(value));
    Handler.wrappers.set(value, {proxy, handler});
  } else {
    ({proxy} = wrapper);
  }
  return proxy;
};

Handler.handler = function(value) {
  var handler, proxy, wrapper;
  if (!(Solid.Utils.isObject(value) && !(value instanceof Element))) {
    return value;
  }
  if (!(wrapper = Handler.wrappers.get(value))) {
    proxy = new Proxy(value, handler = new Handler(value));
    Handler.wrappers.set(value, {proxy, handler});
  } else {
    ({handler} = wrapper);
  }
  return handler;
};

var methods;

methods = {
  trigger: function(property, value, notify) {
    var sub, subs;
    subs = this._child_subscriptions[property];
    if (subs != null ? subs.size : void 0) {
      for (sub of subs) {
        Solid.queueTask(sub, value);
      }
    }
    if (notify) {
      return this.notify(this._target);
    }
  },
  peek: function(property) {
    var value;
    value = this._target[property];
    if (Solid.Utils.isObject(value) && Object.isFrozen(value)) {
      value = Solid.Utils.clone(value);
      this._target[property] = value;
    }
    return Handler$1.wrap(value);
  },
  on: function(property, fn) {
    var base, disposable, handler, value;
    value = this._target[property];
    disposable = null;
    if (Solid.Utils.isObject(value) && !(Solid.Utils.isFunction(value) || value instanceof Element)) {
      handler = Handler$1.handler(value);
      disposable = handler.subscribe(fn);
    }
    (base = this._child_subscriptions)[property] || (base[property] = new Set());
    this._child_subscriptions[property].add(fn);
    return {
      unsubscribe: () => {
        if (disposable) {
          disposable.unsubscribe();
        }
        return this._child_subscriptions[property].delete(fn);
      }
    };
  }
};

Object.assign(Handler$1.prototype, methods);

var methods$1 = methods;

var State, setNested;

setNested = function(item, changes) {
  var handler, isArray, notify, property, value;
  handler = Handler$1.handler(item);
  isArray = Array.isArray(item);
  if (arguments.length === 3) {
    notify = isArray || !(arguments[1] in item);
    if (arguments[2] === void 0) {
      delete item[arguments[1]];
      if (isArray) {
        item.length--;
      }
    } else {
      item[arguments[1]] = arguments[2];
    }
    handler.trigger(arguments[1], arguments[2], notify);
    return;
  }
  for (property in changes) {
    value = changes[property];
    notify = isArray || !(property in item);
    if (value === void 0) {
      delete item[property];
    } else {
      item[property] = value;
    }
    handler.trigger(property, value, notify);
  }
};

var index = State = class State {
  constructor(state) {
    var k, ref;
    Object.defineProperties(this, {
      _target: {
        value: state,
        writable: true
      },
      _disposables: {
        value: [],
        writable: true
      },
      _child_subscriptions: {
        value: {},
        writable: true
      }
    });
    for (k in this._target) {
      this._defineProperty(k);
    }
    if ((ref = Solid.Core.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  set() {
    var args, ref;
    if ((ref = Solid.Core.context) != null ? ref.pure : void 0) {
      return console.log('Cannot update in a Selector');
    }
    args = arguments;
    Solid.run(() => {
      var change, changes, current, i, j, len, property, ref1, ref2, temp, value;
      if (args.length === 1) {
        if (Array.isArray(args[0])) {
          ref1 = args[0];
          for (j = 0, len = ref1.length; j < len; j++) {
            change = ref1[j];
            this.set.apply(this, change);
          }
        } else {
          ref2 = args[0];
          for (property in ref2) {
            value = ref2[property];
            this._setProperty(property, value);
          }
        }
        return;
      }
      current = this._target;
      changes = args[args.length - 1];
      i = 0;
      while (i < args.length - 1 && ((temp = current[args[i]]) != null)) {
        current = temp;
        i++;
      }
      return setNested(current, changes);
    });
  }

  replace() {
    var changes, current, i, property, temp, value;
    if (arguments.length === 1) {
      if (!(arguments[0] instanceof Object)) {
        return console.log('replace must be provided a replacement state');
      }
      changes = arguments[0];
      Solid.run(() => {
        var change, j, len;
        if (!Array.isArray(changes)) {
          changes = Solid.Utils.diff(changes, this._target);
        }
        for (j = 0, len = changes.length; j < len; j++) {
          change = changes[j];
          this.replace.apply(this, change);
        }
      });
    }
    if (arguments.length === 2) {
      this._setProperty(arguments[0], arguments[1]);
      return;
    }
    current = this._target;
    value = arguments[arguments.length - 1];
    property = arguments[arguments.length - 2];
    i = 0;
    while (i < arguments.length - 2 && ((temp = current[arguments[i]]) != null)) {
      current = temp;
      i++;
    }
    setNested(current, property, value);
  }

  select() {
    var j, key, len, selection, selector;
    for (j = 0, len = arguments.length; j < len; j++) {
      selection = arguments[j];
      if (Solid.Utils.isFunction(selection) || 'then' in selection || 'subscribe' in selection) {
        if (!('sid' in selection)) {
          selection = Solid__default(selection);
        }
        this._disposables.push(selection.subscribe((value) => {
          var key, val;
          value = Solid.Utils.unwrap(value, true);
          this.replace([].concat(...((function() {
            var ref, results;
            ref = value || {};
            results = [];
            for (key in ref) {
              val = ref[key];
              results.push(Solid.Utils.diff(val, this._target[key], [key]));
            }
            return results;
          }).call(this))));
        }));
        return;
      }
      for (key in selection) {
        selector = selection[key];
        ((key, selector) => {
          if (!(key in this)) {
            this._defineProperty(key);
          }
          if (!('sid' in selector)) {
            selector = Solid__default(selector);
          }
          this._disposables.push(selector.subscribe((value) => {
            this.replace(Solid.Utils.diff(Solid.Utils.unwrap(value, true), this._target[key], [key]));
          }));
        })(key, selector);
      }
    }
  }

  dispose() {
    var disposable, j, len, ref;
    if (!this._disposables) {
      return;
    }
    ref = this._disposables;
    for (j = 0, len = ref.length; j < len; j++) {
      disposable = ref[j];
      disposable.unsubscribe();
    }
    this._disposables = null;
  }

  _setProperty(property, value) {
    if (!(property in this)) {
      this._defineProperty(property);
    }
    if (value === void 0) {
      delete this._target[property];
    } else {
      this._target[property] = value;
    }
    return this.trigger(property, value);
  }

  _defineProperty(property) {
    return Object.defineProperty(this, property, {
      get: () => {
        var ref, value;
        if (!((ref = Solid.Core.context) != null ? ref.exec : void 0)) {
          return this._target[property];
        }
        if ((value = this._target[property]) != null) {
          if (Solid.Utils.isFunction(value)) {
            return value;
          }
          if (Solid.Utils.isObject(value) && Object.isFrozen(value)) {
            value = Solid.Utils.clone(value);
            this._target[property] = value;
          }
          value = Handler$1.wrap(value);
        }
        Solid.Core.context.disposables.push(this.on(property, Solid.Core.context.exec).unsubscribe);
        return value;
      },
      enumerable: true
    });
  }

};

Object.assign(State.prototype, methods$1);

var DEFINED$1, ImmutableHandler;

DEFINED$1 = {
  '_target': true,
  'on': true,
  'peek': true,
  'map': true
};

var ImmutableHandler$1 = ImmutableHandler = class ImmutableHandler {
  constructor(target, _target, root) {
    this._target = _target;
    this.root = root;
    this.peek = this.peek.bind(this, target);
    this.on = this.on.bind(this, target);
    if (Array.isArray(this._target)) {
      this.map = this.map.bind(this, target);
    }
    this.stateClock = target.clock || Solid.Core.clock;
  }

  get(target, property) {
    var base, base1, ref, value;
    if (this.stateClock < target.clock) {
      this.resolveState(target);
    }
    if (property === 'length' || typeof property === 'symbol') {
      return this._target[property];
    }
    if (property.endsWith('$')) {
      property = property.slice(0, -1);
      if (Solid.Core.context.exec) {
        target[property] || (target[property] = {
          _subs: new Set(),
          clock: Solid.Core.clock
        });
        (base = target[property])._subPath || (base._subPath = new Set());
        target[property]._subPath.add(Solid.Core.context.exec);
        Solid.Core.context.disposables.push(target[property]._subPath.delete(Solid.Core.context.exec));
      }
      return this._target[property];
    }
    if (DEFINED$1[property]) {
      return this[property];
    }
    if (!((ref = Solid.Core.context) != null ? ref.exec : void 0)) {
      return this._target[property];
    }
    if (((value = this._target[property]) != null) && Solid.Utils.isObject(value) && !(value instanceof Element)) {
      if (Solid.Utils.isFunction(value)) {
        return value;
      }
      target[property] || (target[property] = {
        clock: Solid.Core.clock
      });
      (base1 = target[property]).path || (base1.path = target.path.concat([property]));
      value = new Proxy(target[property], new ImmutableHandler(target[property], value, this.root));
    }
    Solid.Core.context.disposables.push(this.on(property, Solid.Core.context.exec).unsubscribe);
    return value;
  }

  set(target, property, value) {
    return true;
  }

  deleteProperty(target, property) {
    return true;
  }

  on(target, property, fn) {
    var base;
    target[property] || (target[property] = {
      clock: Solid.Core.clock
    });
    (base = target[property])._subs || (base._subs = new Set());
    target[property]._subs.add(fn);
    return {
      unsubscribe: function() {
        return target[property]._subs.delete(fn);
      }
    };
  }

  peek(target, property) {
    var base, value;
    if (this.stateClock < target.clock) {
      this.resolveState(target);
    }
    value = this._target[property];
    if ((value == null) || Solid.Utils.isFunction(value) || !Solid.Utils.isObject(value) || (value instanceof Element)) {
      return value;
    }
    target[property] || (target[property] = {});
    (base = target[property]).path || (base.path = target.path.concat([property]));
    return new Proxy(target[property], new ImmutableHandler(target[property], value, this.root));
  }

  map(target, fn) {
    if (this.stateClock < target.clock) {
      this.resolveState(target);
    }
    return this._target.map((value, i, a) => {
      var base;
      if (!((value == null) || Solid.Utils.isFunction(value) || !Solid.Utils.isObject(value) || (value instanceof Element))) {
        target[i] || (target[i] = {});
        (base = target[i]).path || (base.path = target.path.concat([i]));
        value = new Proxy(target[i], new ImmutableHandler(target[i], value, this.root));
      }
      return fn(value, i, a);
    });
  }

  resolveState(target) {
    var current, i, l;
    current = this.root._target;
    i = 0;
    l = target.path.length;
    while (i < l) {
      if (current) {
        current = current[target.path[i]];
      }
      i++;
    }
    this._target = current;
    return this.stateClock = Solid.Core.clock;
  }

};

var ImmutableState,
  indexOf = [].indexOf;

var index$1 = ImmutableState = class ImmutableState {
  constructor(state) {
    var k, ref;
    Object.defineProperties(this, {
      _target: {
        value: state,
        writable: true
      },
      _disposables: {
        value: [],
        writable: true
      },
      _subscriptions: {
        value: {},
        writable: true
      }
    });
    for (k in this._target) {
      this._defineProperty(k);
    }
    if ((ref = Solid.Core.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  set() {
    var args, ref;
    if ((ref = Solid.Core.context) != null ? ref.pure : void 0) {
      return console.log('Cannot update in a Selector');
    }
    args = arguments;
    return Solid.run(() => {
      var change, changes, j, k, l, len, len1, notify, path, property, ref1, ref2, ref3, state, subKeys, subPaths, subs, value;
      if (args.length === 1) {
        if (Array.isArray(args[0])) {
          ref1 = args[0];
          for (j = 0, len = ref1.length; j < len; j++) {
            change = ref1[j];
            this.set.apply(this, change);
          }
        } else {
          ref2 = args[0];
          for (property in ref2) {
            value = ref2[property];
            this._setProperty(property, value);
          }
        }
        return;
      }
      changes = args[args.length - 1];
      ({state, subs, subPaths} = this._resolvePath(args, args.length - 1));
      if (!state) {
        return;
      }
      notify = Array.isArray(state);
      for (property in changes) {
        value = changes[property];
        notify = notify || !(indexOf.call(state, property) >= 0);
        if (value === void 0) {
          delete state[property];
        } else {
          state[property] = value;
        }
        if (subs) {
          this.trigger(subs != null ? (ref3 = subs[property]) != null ? ref3._subs : void 0 : void 0, value);
        }
      }
      if (notify) {
        this.trigger(subs != null ? subs._subs : void 0, state);
      }
      for (l = 0, len1 = subPaths.length; l < len1; l++) {
        path = subPaths[l];
        this.trigger(path.subs, path.state);
      }
      // prune subs
      if (value === void 0) {
        delete subs[property];
      } else if (Array.isArray(value) && !value.length) {
        subKeys = ['_subs', 'clock', '_subPath'];
        for (k in subs[property]) {
          if (!(indexOf.call(subKeys, k) >= 0)) {
            delete subs[property][k];
          }
        }
      }
    });
  }

  replace() {
    var changes, j, len, notify, path, property, ref, state, subPaths, subs, value;
    if (arguments.length === 1) {
      if (!(arguments[0] instanceof Object)) {
        return console.log('replace must be provided a replacement state');
      }
      changes = arguments[0];
      Solid.run(() => {
        var change, j, len;
        if (!Array.isArray(changes)) {
          changes = Solid.Utils.diff(changes, this._target);
        }
        for (j = 0, len = changes.length; j < len; j++) {
          change = changes[j];
          this.replace.apply(this, change);
        }
      });
      return;
    }
    if (arguments.length === 2) {
      this._setProperty(arguments[0], arguments[1]);
      return;
    }
    value = arguments[arguments.length - 1];
    property = arguments[arguments.length - 2];
    ({state, subs, subPaths} = this._resolvePath(arguments, arguments.length - 2));
    if (!state) {
      return;
    }
    if (value === void 0 && Array.isArray(state)) {
      state.length -= 1;
      notify = true;
    } else {
      notify = Array.isArray(state) || !(indexOf.call(state, property) >= 0);
      if (value === void 0) {
        delete state[property];
      } else {
        state[property] = value;
      }
    }
    if (subs) {
      this.trigger((ref = subs[property]) != null ? ref._subs : void 0, value);
    }
    if (notify) {
      this.trigger(subs._subs, state);
    }
    for (j = 0, len = subPaths.length; j < len; j++) {
      path = subPaths[j];
      this.trigger(path.subs, path.state);
    }
  }

  select() {
    var j, key, len, selection, selector;
    for (j = 0, len = arguments.length; j < len; j++) {
      selection = arguments[j];
      if (Solid.Utils.isFunction(selection) || 'then' in selection || 'subscribe' in selection) {
        if (!('sid' in selection)) {
          selection = Solid__default(selection);
        }
        this._disposables.push(selection.subscribe((value) => {
          var key, val;
          value = Solid.Utils.unwrap(value, true);
          this.replace([].concat(...((function() {
            var ref, results;
            ref = value || {};
            results = [];
            for (key in ref) {
              val = ref[key];
              results.push(Solid.Utils.diff(val, this._target[key], [key]));
            }
            return results;
          }).call(this))));
        }));
        return;
      }
      for (key in selection) {
        selector = selection[key];
        ((key, selector) => {
          if (!(key in this)) {
            this._defineProperty(key);
          }
          if (!('sid' in selector)) {
            selector = Solid__default(selector);
          }
          this._disposables.push(selector.subscribe((value) => {
            this.replace(Solid.Utils.diff(Solid.Utils.unwrap(value, true), this._target[key], [key]));
          }));
        })(key, selector);
      }
    }
  }

  on(property, fn) {
    var base, base1;
    (base = this._subscriptions)[property] || (base[property] = {
      clock: Solid.Core.clock
    });
    (base1 = this._subscriptions[property])._subs || (base1._subs = new Set());
    this._subscriptions[property]._subs.add(fn);
    return {
      unsubscribe: () => {
        return this._subscriptions[property]._subs.delete(fn);
      }
    };
  }

  trigger(subs, value) {
    var sub;
    if (!(subs != null ? subs.size : void 0)) {
      return;
    }
    for (sub of subs) {
      Solid.queueTask(sub, value);
    }
  }

  dispose() {
    var disposable, j, len, ref;
    if (!this._disposables) {
      return;
    }
    ref = this._disposables;
    for (j = 0, len = ref.length; j < len; j++) {
      disposable = ref[j];
      disposable.unsubscribe();
    }
    this._disposables = null;
  }

  _resolvePath(path, length) {
    var currentState, currentSubs, i, ref, subPaths;
    currentState = this._target;
    currentSubs = this._subscriptions;
    subPaths = [];
    i = 0;
    while (i < length) {
      if (!currentSubs[path[i]]) {
        currentSubs[path[i]] = {};
      }
      currentSubs = currentSubs[path[i]];
      if (currentState != null) {
        if (currentSubs.clock !== Solid.Core.clock) {
          currentState[path[i]] = Solid.Utils.clone(currentState[path[i]]);
          currentSubs.clock = Solid.Core.clock;
        }
        currentState = currentState[path[i]];
      }
      if (currentSubs != null ? (ref = currentSubs._subPath) != null ? ref.size : void 0 : void 0) {
        subPaths.push({
          state: currentState,
          subs: currentSubs._subPath
        });
      }
      i++;
    }
    return {
      state: currentState,
      subs: currentSubs,
      subPaths
    };
  }

  _setProperty(property, value) {
    var k, ref, ref1, subKeys;
    if (!(property in this)) {
      this._defineProperty(property);
    }
    if (value === void 0) {
      delete this._target[property];
    } else {
      this._target[property] = value;
    }
    this.trigger((ref = this._subscriptions[property]) != null ? ref._subs : void 0, value);
    this.trigger((ref1 = this._subscriptions[property]) != null ? ref1._subPath : void 0, value);
    // prune subs
    if (value === void 0) {
      delete this._subscriptions[property];
    } else if (Array.isArray(value) && !value.length) {
      subKeys = ['_subs', 'clock', '_subPath'];
      for (k in this._subscriptions[property]) {
        if (!(indexOf.call(subKeys, k) >= 0)) {
          delete this._subscriptions[property][k];
        }
      }
    }
  }

  _defineProperty(property) {
    Object.defineProperty(this, property, {
      get: () => {
        var base, base1, ref, value;
        if (!((ref = Solid.Core.context) != null ? ref.exec : void 0)) {
          return this._target[property];
        }
        if (((value = this._target[property]) != null) && Solid.Utils.isObject(value) && !(value instanceof Element)) {
          if (Solid.Utils.isFunction(value)) {
            return value;
          }
          (base = this._subscriptions)[property] || (base[property] = {
            clock: Solid.Core.clock
          });
          (base1 = this._subscriptions[property]).path || (base1.path = [property]);
          value = new Proxy(this._subscriptions[property], new ImmutableHandler$1(this._subscriptions[property], value, this));
        }
        Solid.Core.context.disposables.push(this.on(property, Solid.Core.context.exec).unsubscribe);
        return value;
      },
      enumerable: true
    });
    return Object.defineProperty(this, property + 'Signal', {
      get: () => {
        var base, base1, fn;
        if (fn = Solid.Core.context.exec) {
          (base = this._subscriptions)[property] || (base[property] = {
            _subs: new Set(),
            clock: Solid.Core.clock
          });
          (base1 = this._subscriptions[property])._subPath || (base1._subPath = new Set());
          this._subscriptions[property]._subPath.add(Solid.Core.context.exec);
          Solid.Core.context.disposables.push(() => {
            return this._subscriptions[property]._subPath.delete(fn);
          });
        }
        return this._target[property];
      }
    });
  }

};

exports.State = index;
exports.ImmutableState = index$1;
